/*
 * Copyright © Ricki Hirner (bitfire web engineering).
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 */

package at.bitfire.davdroid.ui

import android.Manifest
import android.accounts.Account
import android.accounts.AccountManager
import android.app.Application
import android.content.ContentResolver
import android.content.ContentUris
import android.content.Intent
import android.content.pm.ApplicationInfo
import android.content.pm.PackageManager
import android.net.ConnectivityManager
import android.os.Build
import android.os.Bundle
import android.os.PowerManager
import android.provider.CalendarContract
import android.provider.ContactsContract
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.Toast
import androidx.activity.viewModels
import androidx.annotation.UiThread
import androidx.appcompat.app.AppCompatActivity
import androidx.core.app.NotificationManagerCompat
import androidx.core.app.ShareCompat
import androidx.core.content.ContextCompat
import androidx.core.content.FileProvider
import androidx.core.content.getSystemService
import androidx.core.content.pm.PackageInfoCompat
import androidx.databinding.DataBindingUtil
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.MutableLiveData
import androidx.lifecycle.Observer
import androidx.lifecycle.viewModelScope
import at.bitfire.dav4jvm.exception.DavException
import at.bitfire.dav4jvm.exception.HttpException
import at.bitfire.davdroid.BuildConfig
import at.bitfire.davdroid.InvalidAccountException
import at.bitfire.davdroid.R
import at.bitfire.davdroid.TextTable
import at.bitfire.davdroid.databinding.ActivityDebugInfoBinding
import at.bitfire.davdroid.log.Logger
import at.bitfire.davdroid.model.AppDatabase
import at.bitfire.davdroid.resource.LocalAddressBook
import at.bitfire.davdroid.settings.AccountSettings
import at.bitfire.davdroid.settings.SettingsManager
import at.bitfire.ical4android.TaskProvider
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import org.apache.commons.io.ByteOrderMark
import org.apache.commons.io.IOUtils
import org.apache.commons.lang3.exception.ExceptionUtils
import org.dmfs.tasks.contract.TaskContract
import java.io.*
import java.util.logging.Level
import java.util.zip.ZipEntry
import java.util.zip.ZipOutputStream

class DebugInfoActivity: AppCompatActivity() {

    companion object {
        /** serialized [Throwable] that causes the problem */
        const val EXTRA_CAUSE = "cause"

        /** logs related to the problem (plain-text [String]) */
        const val EXTRA_LOGS = "logs"

        /** logs related to the problem (path to log file, plain-text [String]) */
        const val EXTRA_LOG_FILE = "logFile"

        /** [android.accounts.Account] (as [android.os.Parcelable]) related to problem */
        const val EXTRA_ACCOUNT = "account"

        /** sync authority name related to problem */
        const val EXTRA_AUTHORITY = "authority"

        /** local resource related to the problem (plain-text [String]) */
        const val EXTRA_LOCAL_RESOURCE = "localResource"

        /** remote resource related to the problem (plain-text [String]) */
        const val EXTRA_REMOTE_RESOURCE = "remoteResource"

        const val FILE_DEBUG_INFO = "debug-info.txt"
        const val FILE_LOGS = "logs.txt"
    }

    private val model by viewModels<ReportModel>()


    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        model.generate(intent.extras)

        val binding = DataBindingUtil.setContentView<ActivityDebugInfoBinding>(this, R.layout.activity_debug_info)
        binding.model = model
        binding.lifecycleOwner = this

        model.cause.observe(this, Observer { cause ->
            if (cause == null)
                return@Observer

            binding.causeCaption.text = when (cause) {
                is HttpException -> getString(if (cause.code / 100 == 5) R.string.debug_info_server_error else R.string.debug_info_http_error)
                is DavException -> getString(R.string.debug_info_webdav_error)
                is IOException, is IOError -> getString(R.string.debug_info_io_error)
                else -> cause::class.java.simpleName
            }

            binding.causeText.text = getString(
                if (cause is HttpException)
                    when {
                        cause.code == 403 -> R.string.debug_info_http_403_description
                        cause.code == 404 -> R.string.debug_info_http_404_description
                        cause.code/100 == 5 -> R.string.debug_info_http_5xx_description
                        else -> R.string.debug_info_unexpected_error
                    }
                else
                    R.string.debug_info_unexpected_error
            )
        })

        model.debugInfo.observe(this, { debugInfo ->
            val showDebugInfo = View.OnClickListener {
                val uri = FileProvider.getUriForFile(this, getString(R.string.authority_debug_provider), debugInfo)
                val intent = Intent(Intent.ACTION_VIEW)
                intent.setDataAndType(uri, "text/plain")
                intent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
                startActivity(Intent.createChooser(intent, null))
            }
            binding.causeView.setOnClickListener(showDebugInfo)
            binding.debugInfoView.setOnClickListener(showDebugInfo)
        })

        model.logFile.observe(this, { logs ->
            binding.logsView.setOnClickListener {
                val uri = FileProvider.getUriForFile(this, getString(R.string.authority_debug_provider), logs)
                val intent = Intent(Intent.ACTION_VIEW)
                intent.setDataAndType(uri, "text/plain")
                intent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
                startActivity(Intent.createChooser(intent, null))
            }
        })
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.activity_debug_info, menu)
        return true
    }


    fun onShare(item: MenuItem? = null) {
        model.generateZip { zipFile ->
            val builder = ShareCompat.IntentBuilder.from(this)
                    .setSubject("${getString(R.string.app_name)} ${BuildConfig.VERSION_NAME} debug info")
                    .setText(getString(R.string.debug_info_attached))
                    .setType("application/zip")
                    .setStream(FileProvider.getUriForFile(this, getString(R.string.authority_debug_provider), zipFile))
            builder.intent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION or Intent.FLAG_ACTIVITY_NEW_TASK)
            builder.startChooser()
        }
    }


    class ReportModel(
            val context: Application
    ): AndroidViewModel(context) {

        private var initialized = false

        val cause = MutableLiveData<Throwable>()
        var logFile = MutableLiveData<File>()
        val localResource = MutableLiveData<String>()
        val remoteResource = MutableLiveData<String>()
        val debugInfo = MutableLiveData<File>()

        val zipFile = MutableLiveData<File>()

        // private storage, not readable by others
        val debugInfoDir = File(context.filesDir, "debug")

        init {
            // create debug info directory
            if (!debugInfoDir.isDirectory && !debugInfoDir.mkdir())
                throw IOException("Couldn't create debug info directory")
        }

        @UiThread
        fun generate(extras: Bundle?) {
            if (initialized)
                return
            initialized = true

            viewModelScope.launch(Dispatchers.Default) {
                val logFileName = extras?.getString(EXTRA_LOG_FILE)
                val logsText = extras?.getString(EXTRA_LOGS)
                if (logFileName != null) {
                    val file = File(logFileName)
                    if (file.isFile && file.canRead())
                        logFile.postValue(file)
                    else
                        Logger.log.warning("Can't read logs from $logFileName")
                } else if (logsText != null) {
                    val file = File(debugInfoDir, FILE_LOGS)
                    if (!file.exists() || file.canWrite()) {
                        file.writer().buffered().use { writer ->
                            IOUtils.copy(StringReader(logsText), writer)
                        }
                        logFile.postValue(file)
                    } else
                        Logger.log.warning("Can't write logs to $file")
                }

                val throwable = extras?.getSerializable(EXTRA_CAUSE) as? Throwable
                cause.postValue(throwable)

                val local = extras?.getString(EXTRA_LOCAL_RESOURCE)
                localResource.postValue(local)

                val remote = extras?.getString(EXTRA_REMOTE_RESOURCE)
                remoteResource.postValue(remote)

                generateDebugInfo(
                        extras?.getParcelable<Account>(EXTRA_ACCOUNT),
                        extras?.getString(EXTRA_AUTHORITY),
                        throwable,
                        local,
                        remote
                )
            }
        }

        fun generateDebugInfo(syncAccount: Account?, syncAuthority: String?, cause: Throwable?, localResource: String?, remoteResource: String?) {
            val debugInfoFile = File(debugInfoDir, FILE_DEBUG_INFO)
            debugInfoFile.writer().buffered().use { writer ->
                writer.append(ByteOrderMark.UTF_BOM)
                writer.append("--- BEGIN DEBUG INFO ---\n\n")

                // begin with most specific information
                if (syncAccount != null || syncAuthority != null) {
                    writer.append("SYNCHRONIZATION INFO\n")
                    if (syncAccount != null)
                        writer.append("Account: $syncAccount\n")
                    if (syncAuthority != null)
                        writer.append("Authority: $syncAuthority\n")
                    writer.append("\n")
                }

                cause?.let {
                    // Log.getStackTraceString(e) returns "" in case of UnknownHostException
                    writer.append("EXCEPTION\n${ExceptionUtils.getStackTrace(cause)}\n")
                }

                // exception details
                if (cause is HttpException) {
                    cause.request?.let { request ->
                        writer.append("HTTP REQUEST\n$request\n")
                        cause.requestBody?.let { writer.append(it) }
                        writer.append("\n\n")
                    }
                    cause.response?.let { response ->
                        writer.append("HTTP RESPONSE\n$response\n")
                        cause.responseBody?.let { writer.append(it) }
                        writer.append("\n")
                    }
                }

                if (localResource != null)
                    writer.append("LOCAL RESOURCE\n$localResource\n\n")

                if (remoteResource != null)
                    writer.append("REMOTE RESOURCE\n$remoteResource\n\n")

                // software info
                try {
                    writer.append("SOFTWARE INFORMATION\n")
                    val table = TextTable("Package", "Version", "Code", "Installer", "Notes")
                    val pm = context.packageManager
                    val appIDs = mutableSetOf(      // we always want info about these packages
                            BuildConfig.APPLICATION_ID,                         // DAVx5
                            TaskProvider.ProviderName.OpenTasks.packageName,    // OpenTasks
                            TaskProvider.ProviderName.TasksOrg.packageName      // tasks.org
                    )
                    // add info about contact, calendar, task provider
                    for (authority in arrayOf(ContactsContract.AUTHORITY, CalendarContract.AUTHORITY, TaskProvider.ProviderName.OpenTasks.authority))
                        pm.resolveContentProvider(authority, 0)?.let { appIDs += it.packageName }
                    // add info about available contact, calendar, task apps
                    for (uri in arrayOf(ContactsContract.Contacts.CONTENT_URI, CalendarContract.Events.CONTENT_URI, TaskContract.Tasks.getContentUri(TaskProvider.ProviderName.OpenTasks.authority))) {
                        val viewIntent = Intent(Intent.ACTION_VIEW, ContentUris.withAppendedId(uri, 1))
                        for (info in pm.queryIntentActivities(viewIntent, 0))
                            appIDs += info.activityInfo.packageName
                    }

                    for (appID in appIDs)
                        try {
                            val info = pm.getPackageInfo(appID, 0)
                            val appInfo = info.applicationInfo
                            val notes = mutableListOf<String>()
                            if (!appInfo.enabled)
                                notes += "disabled"
                            if (appInfo.flags.and(ApplicationInfo.FLAG_EXTERNAL_STORAGE) != 0)
                                notes += "<em>on external storage</em>"
                            table.addLine(appID, info.versionName, PackageInfoCompat.getLongVersionCode(info), pm.getInstallerPackageName(appID) ?: '—', notes.joinToString(", "))
                        } catch(e: PackageManager.NameNotFoundException) {
                        }
                    writer.append(table.toString())
                } catch(e: Exception) {
                    Logger.log.log(Level.SEVERE, "Couldn't get software information", e)
                }

                // system info
                writer.append(
                        "SYSTEM INFORMATION\n\n" +
                        "Android version: ${Build.VERSION.RELEASE} (${Build.DISPLAY})\n" +
                        "Device: ${Build.MANUFACTURER} ${Build.MODEL} (${Build.DEVICE})\n\n"
                )

                // connectivity
                context.getSystemService<ConnectivityManager>()?.let { connectivityManager ->
                    writer.append("\nCONNECTVITY\n\n")
                    val activeNetwork = if (Build.VERSION.SDK_INT >= 23) connectivityManager.activeNetwork else null
                    connectivityManager.allNetworks.sortedByDescending { it == activeNetwork }.forEach { network ->
                        val capabilities = connectivityManager.getNetworkCapabilities(network)!!
                        writer  .append(if (network == activeNetwork) " ☒ " else " ☐ ")
                                .append(capabilities.toString().replace('&',' '))
                                .append('\n')
                    }
                    writer.append('\n')

                    if (Build.VERSION.SDK_INT >= 23)
                        connectivityManager.defaultProxy?.let { proxy ->
                            writer.append("System default proxy: ${proxy.host}:${proxy.port}\n")
                        }
                    if (Build.VERSION.SDK_INT >= 24)
                        writer.append("Data saver: ").append(when (connectivityManager.restrictBackgroundStatus) {
                            ConnectivityManager.RESTRICT_BACKGROUND_STATUS_ENABLED -> "enabled"
                            ConnectivityManager.RESTRICT_BACKGROUND_STATUS_WHITELISTED -> "whitelisted"
                            ConnectivityManager.RESTRICT_BACKGROUND_STATUS_DISABLED -> "disabled"
                            else -> connectivityManager.restrictBackgroundStatus.toString()
                        }).append('\n')
                    writer.append('\n')
                }

                writer.append("\nCONFIGURATION\n\n")
                // power saving
                if (Build.VERSION.SDK_INT >= 23)
                    context.getSystemService<PowerManager>()?.let { powerManager ->
                        writer.append("Power saving disabled: ")
                                .append(if (powerManager.isIgnoringBatteryOptimizations(BuildConfig.APPLICATION_ID)) "yes" else "no")
                                .append('\n')
                    }
                // system-wide sync
                writer  .append("System-wide synchronization: ")
                        .append(if (ContentResolver.getMasterSyncAutomatically()) "automatically" else "manually")
                        .append('\n')
                // notifications
                val nm = NotificationManagerCompat.from(context)
                writer.append("\nNotifications")
                if (!nm.areNotificationsEnabled())
                    writer.append(" (blocked!)")
                writer.append(":\n")
                if (Build.VERSION.SDK_INT >= 26) {
                    val channelsWithoutGroup = nm.notificationChannels.toMutableSet()
                    for (group in nm.notificationChannelGroups) {
                        writer.append(" - ${group.id} ")
                        if (Build.VERSION.SDK_INT >= 28)
                            writer.append(" isBlocked=${group.isBlocked}")
                        writer.append('\n')
                        for (channel in group.channels) {
                            writer.append("  * ${channel.id}: importance=${channel.importance}\n")
                            channelsWithoutGroup -= channel
                        }
                    }
                    for (channel in channelsWithoutGroup)
                        writer.append(" - ${channel.id}: importance=${channel.importance}\n")
                }
                writer.append('\n')
                // permissions
                writer.append("Permissions:\n")
                for (permission in arrayOf(Manifest.permission.READ_CONTACTS, Manifest.permission.WRITE_CONTACTS,
                        Manifest.permission.READ_CALENDAR, Manifest.permission.WRITE_CALENDAR,
                        TaskProvider.PERMISSION_OPENTASKS_READ, TaskProvider.PERMISSION_OPENTASKS_WRITE,
                        TaskProvider.PERMISSION_TASKS_ORG_READ, TaskProvider.PERMISSION_TASKS_ORG_WRITE,
                        Manifest.permission.ACCESS_COARSE_LOCATION)) {
                    val shortPermission = permission.replace(Regex("^.+\\.permission\\."), "")
                    writer  .append(" - $shortPermission: ")
                            .append(if (ContextCompat.checkSelfPermission(context, permission) == PackageManager.PERMISSION_GRANTED)
                                "granted"
                            else
                                "denied")
                            .append('\n')
                }
                writer.append('\n')

                writer.append("\nACCOUNTS\n\n")
                // main accounts
                val accountManager = AccountManager.get(context)
                val mainAccounts = accountManager.getAccountsByType(context.getString(R.string.account_type))
                val addressBookAccounts = accountManager.getAccountsByType(context.getString(R.string.account_type_address_book)).toMutableList()
                for (account in mainAccounts) {
                    dumpMainAccount(account, writer)

                    val iter = addressBookAccounts.iterator()
                    while (iter.hasNext()) {
                        val addressBookAccount = iter.next()
                        val mainAccount = Account(
                                accountManager.getUserData(addressBookAccount, LocalAddressBook.USER_DATA_MAIN_ACCOUNT_NAME),
                                accountManager.getUserData(addressBookAccount, LocalAddressBook.USER_DATA_MAIN_ACCOUNT_TYPE)
                        )
                        if (mainAccount == account) {
                            dumpAddressBookAccount(addressBookAccount, accountManager, writer)
                            iter.remove()
                        }
                    }
                }
                if (addressBookAccounts.isNotEmpty()) {
                    writer.append("Address book accounts without main account:\n")
                    for (account in addressBookAccounts)
                        dumpAddressBookAccount(account, accountManager, writer)
                }

                // database dump
                writer.append("\nDATABASE DUMP\n\n")
                AppDatabase.getInstance(context).dump(writer)

                // app settings
                writer.append("\nAPP SETTINGS\n\n")
                SettingsManager.getInstance(context).dump(writer)

                writer.append("--- END DEBUG INFO ---\n")
                writer.toString()
            }
            debugInfo.postValue(debugInfoFile)
        }

        fun generateZip(callback: (File) -> Unit) {
            val context = getApplication<Application>()

            viewModelScope.launch(Dispatchers.IO) {
                try {
                    val zipFile = File(debugInfoDir, "davx5-debug.zip")
                    Logger.log.fine("Writing debug info to ${zipFile.absolutePath}")
                    ZipOutputStream(zipFile.outputStream().buffered()).use { zip ->
                        debugInfo.value?.let { debugInfo ->
                            zip.putNextEntry(ZipEntry("debug-info.txt"))
                            debugInfo.inputStream().use {
                                IOUtils.copy(it, zip)
                            }
                            zip.closeEntry()
                        }

                        // logs
                        logFile.value?.let { logs ->
                            zip.putNextEntry(ZipEntry(logs.name))
                            logs.inputStream().use {
                                IOUtils.copy(it, zip)
                            }
                            zip.closeEntry()
                        }
                    }

                    withContext(Dispatchers.Main) {
                        callback(zipFile)
                    }
                } catch(e: IOException) {
                    // creating attachment with debug info failed
                    Logger.log.log(Level.SEVERE, "Couldn't attach debug info", e)
                    Toast.makeText(context, e.toString(), Toast.LENGTH_LONG).show()
                }
            }
        }


        private fun dumpMainAccount(account: Account, writer: Writer) {
            val context = getApplication<Application>()

            writer.append(" - Account: ${account.name}\n")
            val table = TextTable("Authority", "isSyncable", "getSyncAutomatically", "Sync interval")

            for (authority in arrayOf(
                    context.getString(R.string.address_books_authority),
                    CalendarContract.AUTHORITY,
                    ContactsContract.AUTHORITY,     // Should never be set! Main accounts shall not contain contacts.
                    TaskProvider.ProviderName.OpenTasks.authority
            ))
                table.addLine(
                        authority,
                        ContentResolver.getIsSyncable(account, authority),
                        ContentResolver.getSyncAutomatically(account, authority),
                        ContentResolver.getPeriodicSyncs(account, authority).firstOrNull()?.let { periodicSync ->
                            "${periodicSync.period/60} min"
                        }
                )
            writer.append(table.toString())

            try {
                val accountSettings = AccountSettings(context, account)
                writer.append("WiFi only: ${accountSettings.getSyncWifiOnly()}")
                accountSettings.getSyncWifiOnlySSIDs()?.let { ssids ->
                    writer.append(", SSIDs: ${ssids.joinToString(", ")}")
                }
                writer.append("\nContact group method: ${accountSettings.getGroupMethod()}\n" +
                        "Time range (past days): ${accountSettings.getTimeRangePastDays()}\n" +
                        "Default alarm (min before): ${accountSettings.getDefaultAlarm()}\n" +
                        "Manage calendar colors: ${accountSettings.getManageCalendarColors()}\n" +
                        "Use event colors: ${accountSettings.getEventColors()}\n")
            } catch(e: InvalidAccountException) {
                writer.append("$e\n")
            }
            writer.append('\n')
        }

        private fun dumpAddressBookAccount(account: Account, accountManager: AccountManager, writer: Writer) {
            writer.append("   * Address book: ${account.name}\n")
            val table = TextTable("isSyncable", "getSyncAutomatically", "Sync interval")
            table.addLine(
                    ContentResolver.getIsSyncable(account, ContactsContract.AUTHORITY),
                    ContentResolver.getSyncAutomatically(account, ContactsContract.AUTHORITY),
                    ContentResolver.getPeriodicSyncs(account, ContactsContract.AUTHORITY).firstOrNull()?.let { periodicSync ->
                        "${periodicSync.period/60} min"
                    }
            )
            writer  .append(TextTable.indent(table.toString(), 3))
                    .append("   URL: ${accountManager.getUserData(account, LocalAddressBook.USER_DATA_URL)}\n")
                    .append("   Read-only: ${accountManager.getUserData(account, LocalAddressBook.USER_DATA_READ_ONLY) ?: 0}\n\n")
        }

    }

}
